from citrination_client.models import Event
from citrination_client.base.errors import CitrinationClientError

class ServiceStatus(object):
    """
    The status of a service for a given model on Citrination. The status of a
    service is summarized by the ready value indicating whether or not the service
    is in a state where it can be utilized. It also includes more detailed
    information about the current state of the service in the form of a descriptive
    sentence, a context string and an event object.
    """

    @staticmethod
    def from_response_dict(response_dict):

        if "event" in response_dict:
            event_dict = response_dict["event"]
            event = Event(
                title = event_dict["title"],
                subtitle = event_dict["subtitle"],
                normalized_progress = event_dict["normalizedProgress"]
            )

            if "subevent" in response_dict["event"]:
                subevent_dict = response_dict["event"]["subevent"]
                subevent = Event(
                    title = subevent_dict["title"],
                    subtitle = subevent_dict["subtitle"],
                    normalized_progress = subevent_dict["normalizedProgress"]
                )
                event.subevent = subevent
        else:
            event = None

        return ServiceStatus(
            ready = response_dict["ready"],
            reason = response_dict["reason"],
            context = response_dict["context"],
            event = event
        )

    def __init__(self, ready, context, reason, event):
        """
        Constructor.

        :param ready: Boolean indicating whether or not the service can be used
        :type ready: bool
        :param context: A contextual description of the current status: "notice",
            "success", "error"
        :type context: str
        :param reason: A full sentence explanation of the service's status
        :type reason: str
        :param event: An event object describing the current state of the service's
            progress toward readiness
        :type event: Event
        """
        self._ready   = ready
        self._context = context
        self._event   = event
        self._reason  = reason
        self._validate_service_status()

    @property
    def ready(self):
        return self._ready

    @ready.setter
    def ready(self, value):
        self._ready = value

    @ready.deleter
    def ready(self):
        self._ready = None

    @property
    def context(self):
        return self._context

    @context.setter
    def context(self, value):
        self._context = value

    @context.deleter
    def context(self):
        self._context = None

    @property
    def event(self):
        return self._event

    @event.setter
    def event(self, value):
        self._event = value

    @event.deleter
    def event(self):
        self._event = None

    @property
    def reason(self):
        return self._reason

    @reason.setter
    def reason(self, value):
        self._reason = value

    @reason.deleter
    def reason(self):
        self._reason = None

    def is_ready(self):
        """
        Indicates whether or not the service is ready to be used.

        :return: A boolean
        :rtype: bool
        """
        return self.ready == True

    def _validate_service_status(self):
        if self.ready and self.event:
            if self.event.normalized_progress < 1.0:
                raise CitrinationClientError("Events still in progress but ready status set to True")




from pypif.util.serializable import Serializable

from citrination_client.search.core.query.data_query import DataQuery


class DataScope(Serializable):
    """
    Query to against data.
    """

    def __init__(self, query=None, **kwargs):
        """
        Constructor.

        :param query: One or more :class:`DataQuery` objects with the queries to run.
        """
        self._query = None
        self.query = query

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, query):
        self._query = self._get_object(DataQuery, query)

    @query.deleter
    def query(self):
        self._query = None



from citrination_client.search.query_encoder import QueryEncoder
from citrination_client.search import *
from citrination_client.search import routes as routes
from citrination_client.util import config as client_config
from citrination_client.base.base_client import BaseClient
from citrination_client.base.errors import RequestTimeoutException
from citrination_client.base.errors import CitrinationClientError

from pypif.util.case import to_camel_case
from pypif.util.case import keys_to_snake_case

from copy import deepcopy
import json
import requests

DEFAULT_FAILURE_MESSAGE = "An error occurred requesting search results from Citrination"
MAX_QUERY_DEPTH = 50000


class SearchClient(BaseClient):
    def __init__(self, api_key, webserver_host="https://citrination.com", suppress_warnings=False):
        members = [
            "pif_search",
            "pif_multi_search",
            "dataset_search"
        ]
        super(SearchClient, self).__init__(api_key, webserver_host, members, suppress_warnings=suppress_warnings)

    def _handle_response(self, response, failure_message=DEFAULT_FAILURE_MESSAGE):
        if response.status_code == 204:
            raise RequestTimeoutException()

        return super(SearchClient, self)._handle_response(response, failure_message)

    def _validate_search_query(self, returning_query):
        """
        Checks to see that the query will not exceed the max query depth

        :param returning_query: The PIF system or Dataset query to execute.
        :type returning_query: :class:`PifSystemReturningQuery` or :class: `DatasetReturningQuery`
        """

        start_index = returning_query.from_index or 0
        size = returning_query.size or 0

        if start_index < 0:
            raise CitrinationClientError(
                "start_index cannot be negative. Please enter a value greater than or equal to zero")
        if size < 0:
            raise CitrinationClientError("Size cannot be negative. Please enter a value greater than or equal to zero")
        if start_index + size > MAX_QUERY_DEPTH:
            raise CitrinationClientError(
                "Citrination does not support pagination past the {0}th result. Please reduce either the from_index and/or size such that their sum is below {0}".format(
                    MAX_QUERY_DEPTH))

    def pif_search(self, pif_system_returning_query):
        """
        Run a PIF query against Citrination.

        :param pif_system_returning_query: The PIF system query to execute.
        :type pif_system_returning_query: :class:`PifSystemReturningQuery`
        :return: :class:`PifSearchResult` object with the results of the query.
        :rtype: :class:`PifSearchResult`
        """

        self._validate_search_query(pif_system_returning_query)
        return self._execute_search_query(
            pif_system_returning_query,
            PifSearchResult
        )

    def dataset_search(self, dataset_returning_query):
        """
        Run a dataset query against Citrination.

        :param dataset_returning_query: :class:`DatasetReturningQuery` to execute.
        :type dataset_returning_query: :class:`DatasetReturningQuery`
        :return: Dataset search result object with the results of the query.
        :rtype: :class:`DatasetSearchResult`
        """

        self._validate_search_query(dataset_returning_query)
        return self._execute_search_query(
            dataset_returning_query,
            DatasetSearchResult
        )

    def _execute_search_query(self, returning_query, result_class):
        """
        Run a PIF query against Citrination.

        :param returning_query: :class:`BaseReturningQuery` to execute.
        :param result_class: The class of the result to return.
        :return: ``result_class`` object with the results of the query.
        """
        if returning_query.from_index:
            from_index = returning_query.from_index
        else:
            from_index = 0

        if returning_query.size != None:
            size = min(returning_query.size, client_config.max_query_size)
        else:
            size = client_config.max_query_size

        if (size == client_config.max_query_size and
                    size != returning_query.size):
            self._warn("Query size greater than max system size - only {} results will be returned".format(size))

        time = 0.0;
        hits = [];
        while True:
            sub_query = deepcopy(returning_query)
            sub_query.from_index = from_index + len(hits)
            partial_results = self._search_internal(sub_query, result_class)
            total = partial_results.total_num_hits
            time += partial_results.took
            if partial_results.hits is not None:
                hits.extend(partial_results.hits)
            if len(hits) >= size or len(hits) >= total or sub_query.from_index >= total:
                break

        return result_class(hits=hits, total_num_hits=total, took=time)

    def _search_internal(self, returning_query, result_class):
        if result_class == PifSearchResult:
            route = routes.pif_search
            failure_message = "Error while making PIF search request"

        elif result_class == DatasetSearchResult:
            route = routes.dataset_search
            failure_message = "Error while making dataset search request"

        response_json = self._get_success_json(self._post(
            route, data=json.dumps(returning_query, cls=QueryEncoder),
            failure_message=failure_message))

        return result_class(**keys_to_snake_case(response_json['results']))

    def pif_multi_search(self, multi_query):
        """
        Run each in a list of PIF queries against Citrination.

        :param multi_query: :class:`MultiQuery` object to execute.
        :return: :class:`PifMultiSearchResult` object with the results of the query.
        """
        failure_message = "Error while making PIF multi search request"
        response_dict = self._get_success_json(
            self._post(routes.pif_multi_search, data=json.dumps(multi_query, cls=QueryEncoder),
                       failure_message=failure_message))

        return PifMultiSearchResult(**keys_to_snake_case(response_dict['results']))

    def generate_simple_chemical_query(self, name=None, chemical_formula=None, property_name=None, property_value=None,
                                       property_min=None, property_max=None, property_units=None, reference_doi=None,
                                       include_datasets=[], exclude_datasets=[], from_index=None, size=None):
        """
        This method generates a :class:`PifSystemReturningQuery` object using the
        supplied arguments. All arguments that accept lists have logical OR's on the queries that they generate.
        This means that, for example, simple_chemical_search(name=['A', 'B']) will match records that have name
        equal to 'A' or 'B'.

        Results will be pulled into the extracted field of the :class:`PifSearchHit` objects that are returned. The
        name will appear under the key "name", chemical formula under "chemical_formula", property name under
        "property_name", value of the property under "property_value", units of the property under "property_units",
        and reference DOI under "reference_doi".

        This method is only meant for execution of very simple queries. More complex queries must use the search method
        that accepts a :class:`PifSystemReturningQuery` object.

        :param name: One or more strings with the names of the chemical system to match.
        :type name: str or list of str
        :param chemical_formula:  One or more strings with the chemical formulas to match.
        :type chemical_formula: str or list of str
        :param property_name: One or more strings with the names of the property to match.
        :type property_name: str or list of str
        :param property_value: One or more strings or numbers with the exact values to match.
        :type property_value: str or int or float or list of str or int or float
        :param property_min: A single string or number with the minimum value to match.
        :type property_min: str or int or float
        :param property_max: A single string or number with the maximum value to match.
        :type property_max: str or int or float
        :param property_units: One or more strings with the property units to match.
        :type property_units: str or list of str
        :param reference_doi: One or more strings with the DOI to match.
        :type reference_doin: str or list of str
        :param include_datasets: One or more integers with dataset IDs to match.
        :type include_datasets: int or list of int
        :param exclude_datasets: One or more integers with dataset IDs that must not match.
        :type exclude_datasets: int or list of int
        :param from_index: Index of the first record to match.
        :type from_index: int
        :param size: Total number of records to return.
        :type size: int
        :return: A query to to be submitted with the pif_search method
        :rtype: :class:`PifSystemReturningQuery`
        """
        pif_system_query = PifSystemQuery()
        pif_system_query.names = FieldQuery(
            extract_as='name',
            filter=[Filter(equal=i) for i in self._get_list(name)])
        pif_system_query.chemical_formula = ChemicalFieldQuery(
            extract_as='chemical_formula',
            filter=[ChemicalFilter(equal=i) for i in self._get_list(chemical_formula)])
        pif_system_query.references = ReferenceQuery(doi=FieldQuery(
            extract_as='reference_doi',
            filter=[Filter(equal=i) for i in self._get_list(reference_doi)]))

        # Generate the parts of the property query
        property_name_query = FieldQuery(
            extract_as='property_name',
            filter=[Filter(equal=i) for i in self._get_list(property_name)])
        property_units_query = FieldQuery(
            extract_as='property_units',
            filter=[Filter(equal=i) for i in self._get_list(property_units)])
        property_value_query = FieldQuery(
            extract_as='property_value',
            filter=[])
        for i in self._get_list(property_value):
            property_value_query.filter.append(Filter(equal=i))
        if property_min is not None or property_max is not None:
            property_value_query.filter.append(Filter(min=property_min, max=property_max))

        # Generate the full property query
        pif_system_query.properties = PropertyQuery(
            name=property_name_query,
            value=property_value_query,
            units=property_units_query)

        # Generate the dataset query
        dataset_query = list()
        if include_datasets:
            dataset_query.append(DatasetQuery(logic='MUST', id=[Filter(equal=i) for i in include_datasets]))
        if exclude_datasets:
            dataset_query.append(DatasetQuery(logic='MUST_NOT', id=[Filter(equal=i) for i in exclude_datasets]))

        # Run the query
        pif_system_returning_query = PifSystemReturningQuery(
            query=DataQuery(
                system=pif_system_query,
                dataset=dataset_query),
            from_index=from_index,
            size=size,
            score_relevance=True)

        return pif_system_returning_query

    @staticmethod
    def _get_list(values):
        """
        Helper method that wraps values in a list. If the input is a list then it is returned. If the input is None then an empty list is returned. For anything else, the input value is wrapped as a single-element list.
        
        :param values: Value to make sure exists in a list.
        :return: List with the input values.
        """
        if values is None:
            return []
        elif isinstance(values, list):
            return values
        else:
            return [values]



from pypif.util.serializable import Serializable


class Filter(Serializable):
    """
    Filter that can be applied to any field.
    """

    def __init__(self, logic=None, weight=None, exists=None, equal=None, min=None, max=None, exact=None,
                 filter=None, **kwargs):
        """
        Constructor.

        :param logic: Logic for this filter. Must be equal to one of "MUST", "MUST_NOT", "SHOULD", or "OPTIONAL".
        :param weight: Weight for the filter.
        :param exists: True/False to simply test whether the field exists and has a non-null value.
        :param equal: String with the phrase to match against.
        :param min: String with the minimum value of a range to match against.
        :param max: String with the maximum value of a range to match against.
        :param exact: True/False to set whether the "equal" filter should be an exact match.
        :param filter: List of :class:`Filter` objects with sub-filters.
        """
        self._logic = None
        self.logic = logic
        self._weight = None
        self.weight = weight
        self._exists = None
        self.exists = exists
        self._equal = None
        self.equal = equal
        self._min = None
        self.min = min
        self._max = None
        self.max = max
        self._exact = None
        self.exact = exact
        self._filter = None
        self.filter = filter

    @property
    def logic(self):
        return self._logic

    @logic.setter
    def logic(self, logic):
        self._logic = logic

    @logic.deleter
    def logic(self):
        self._logic = None

    @property
    def weight(self):
        return self._weight

    @weight.setter
    def weight(self, weight):
        self._weight = weight

    @weight.deleter
    def weight(self):
        self._weight = None

    @property
    def exists(self):
        return self._exists

    @exists.setter
    def exists(self, exists):
        self._exists = exists

    @exists.deleter
    def exists(self):
        self._exists = None

    @property
    def equal(self):
        return self._equal

    @equal.setter
    def equal(self, equal):
        self._equal = equal

    @equal.deleter
    def equal(self):
        self._equal = None

    @property
    def min(self):
        return self._min

    @min.setter
    def min(self, min):
        self._min = min

    @min.deleter
    def min(self):
        self._min = None

    @property
    def max(self):
        return self._max

    @max.setter
    def max(self, max):
        self._max = max

    @max.deleter
    def max(self):
        self._max = None

    @property
    def equal(self):
        return self._equal

    @equal.setter
    def equal(self, equal):
        self._equal = equal

    @equal.deleter
    def equal(self):
        self._equal = None

    @property
    def exact(self):
        return self._exact

    @exact.setter
    def exact(self, exact):
        self._exact = exact

    @exact.deleter
    def exact(self):
        self._exact = None

    @property
    def filter(self):
        return self._filter

    @filter.setter
    def filter(self, filter):
        self._filter = self._get_object(Filter, filter)

    @filter.deleter
    def filter(self):
        self._filter = None



from citrination_client.models.columns.base import BaseColumn
from citrination_client.base.errors import CitrinationClientError

from six import string_types

class CategoricalColumn(BaseColumn):
    """
    A categorical column configuration for a data view. Parameterized
    with the basic column options and a list of valid values for the
    column to have.
    """

    TYPE = "Categorical"

    def __init__(self, name, role, group_by_key=False, units=None, categories=[]):
        """
        Constructor.

        :param name: The name of the column
        :type name: str
        :param role: The role the column will play in machine learning:
                       "Input"
                       "Output"
                       "Latent Variable"
                       "Ignore"
        :type role: str
        :param group_by_key: Whether or not this column should be used for
            grouping during cross validation
        :type group_by_key: bool
        :param units: Optionally, the units for the column
        :type units: str
        :param categories: An array of strings that are valid values for data
            in this column
        :type categories: list of str
        """
        super(CategoricalColumn, self).__init__(name=name,
                                                role=role,
                                                group_by_key=group_by_key,
                                                units=units)
        self.categories = categories

    def build_options(self):
        return {
            "categories": self.categories
        }

    @property
    def categories(self):
        return self._categories

    @categories.setter
    def categories(self, value):
        self._validate_categories(value)
        self._categories = [str(v) for v in value]

    @categories.deleter
    def categories(self):
        self._categories = None

    def _validate_categories(self, categories):
        if type(categories) is not list:
            raise CitrinationClientError("CategoricalColumn requires that the categories value is a list of strings")

        if not all(isinstance(item, string_types) for item in categories):
            raise CitrinationClientError("CategoricalColumn requires that the categories value is a list of strings")



from citrination_client.data.dataset import Dataset
from citrination_client.data.dataset_file import DatasetFile
from citrination_client.data.upload_result import UploadResult
from citrination_client.data.dataset_version import DatasetVersion
from citrination_client.data.client import DataClient


class Event(object):
    """
    The representation of an in progress process. Described by a title, subtitle and subevent.
    """

    def __init__(self, title, subtitle, normalized_progress, subevent=None):
        """
        Constructor.

        :param title: The title of the event
        :type title: str
        :param subtitle: More detail about the event
        :type subtitle: str
        :param subevent: An event object describing the current state of the service's
            progress toward readiness
        :type subevent: Event
        :param normalized_progress: The fractional representation of the status of the event
        :type normalized_progress: float
        """
        self._title = title
        self._subtitle = subtitle
        self._subevent = subevent
        self._normalized_progress = normalized_progress

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, value):
        self._title = value

    @title.deleter
    def title(self):
        self._title = None

    @property
    def subtitle(self):
        return self._subtitle

    @subtitle.setter
    def subtitle(self, value):
        self._subtitle = value

    @subtitle.deleter
    def subtitle(self):
        self._subtitle = None

    @property
    def subevent(self):
        return self._subevent

    @subevent.setter
    def subevent(self, value):
        self._subevent = value

    @subevent.deleter
    def subevent(self):
        self._subevent = None

    @property
    def normalized_progress(self):
        return self._normalized_progress

    @normalized_progress.setter
    def normalized_progress(self, value):
        self._normalized_progress = value

    @normalized_progress.deleter
    def normalized_progress(self):
        self._normalized_progress = None



from citrination_client.models.columns.base import BaseColumn
from citrination_client.base.errors import CitrinationClientError

class VectorColumn(BaseColumn):
    """
    An vector column configuration for a data view.
    Parameterized with the basic column options and a length parameter expressing the length of vectors in this column.
    """

    TYPE = "Vector"

    def __init__(self, name, role, group_by_key=False, units=None, length=None):
        """
        Constructor.

        :param name: The name of the column
        :type name: str
        :param role: The role the column will play in machine learning:
                       "Input"
                       "Output"
                       "Latent Variable"
                       "Ignore"
        :type role: str
        :param group_by_key: Whether or not this column should be used for
            grouping during cross validation
        :type group_by_key: bool
        :param units: Optionally, the units for this column
        :type units: str
        :param length: The length of vectors in this column
        :type length: int
        """
        super(VectorColumn, self).__init__(name=name,
                                           role=role,
                                           group_by_key=group_by_key,
                                           units=units)
        self.length = length

    def build_options(self):
        return {
            "length": self.length
        }

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, value):
        try:
            self._length = int(value)
        except ValueError:
            raise CitrinationClientError("When constructing a VectorColumn, parameter length must be castable as an int")

    @length.deleter
    def length(self):
        self._length = None



from citrination_client.search.pif.query.core.field_query import FieldQuery
from citrination_client.search.pif.query.core.reference_query import ReferenceQuery
from citrination_client.search.pif.query.core.value_query import ValueQuery


class PropertyQuery(ValueQuery):
    """
    Class to query against a single property.
    """

    def __init__(self, logic=None, weight=None, simple=None, simple_weight=None, extract_as=None, extract_all=None,
                 extract_when_missing=None, tags=None, length=None, offset=None, conditions=None, data_type=None,
                 name=None, value=None, file=None, units=None, references=None, query=None, **kwargs):
        """
        Constructor.

        :param logic: Logic for this filter. Must be equal to one of "MUST", "MUST_NOT", "SHOULD", or "OPTIONAL".
        :param weight: Weight of the query.
        :param simple: String with the simple query to run against all fields.
        :param simple_weight: Dictionary of relative paths to their weights for simple queries.
        :param extract_as: String with the alias to save this field under.
        :param extract_all: Boolean setting whether all values in an array should be extracted.
        :param extract_when_missing: Any valid JSON-supported object or PIF object. This value is returned when a value is missing that should be extracted (and the overall query is still satisfied).
        :param tags: One or more :class:`FieldQuery` operations against the tags field.
        :param length: One or more :class:`FieldQuery` operations against the length field.
        :param offset: One or more :class:`FieldQuery` operations against the offset field.
        :param conditions: One or more :class:`ValueQuery` operations against the conditions.
        :param data_type: One or more :class:`FieldQuery` operations against the dataType field.
        :param name: One or more :class:`FieldQuery` operations against the name field.
        :param value: One or more :class:`FieldQuery` operations against the value.
        :param file: One or more :class:`FileReferenceQuery` operations against the file.
        :param units: One or more :class:`FieldQuery` operations against the units field.
        :param references: One or more :class:`ReferenceQuery` operations against the references field.
        :param query: One or more :class:`PropertyQuery` objects with nested queries.
        """
        super(PropertyQuery, self).__init__(
            logic=logic, weight=weight, simple=simple, simple_weight=simple_weight, extract_as=extract_as,
            extract_all=extract_all, extract_when_missing=extract_when_missing, tags=tags, length=length,
            offset=offset, name=name, value=value, file=file, units=units, **kwargs)
        self._conditions = None
        self.conditions = conditions
        self._data_type = None
        self.data_type = data_type
        self._references = None
        self.references = references
        self._query = None
        self.query = query

    @property
    def conditions(self):
        return self._conditions

    @conditions.setter
    def conditions(self, conditions):
        self._conditions = self._get_object(ValueQuery, conditions)

    @conditions.deleter
    def conditions(self):
        self._conditions = None

    @property
    def data_type(self):
        return self._data_type

    @data_type.setter
    def data_type(self, data_type):
        self._data_type = self._get_object(FieldQuery, data_type)

    @data_type.deleter
    def data_type(self):
        self._data_type = None

    @property
    def references(self):
        return self._references

    @references.setter
    def references(self, references):
        self._references = self._get_object(ReferenceQuery, references)

    @references.deleter
    def references(self):
        self._references = None

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, query):
        self._query = self._get_object(PropertyQuery, query)

    @query.deleter
    def query(self):
        self._query = None



class DataViewStatus(object):
    """
    A summary of each of the services a data view exposes.
    """

    def __init__(self, predict=None, experimental_design=None, data_reports=None, model_reports=None):
        """
        Constructor.

        :param predict: The status predict
        :type predict: ServiceStatus
        :param experimental_design: The status of experimental_design
        :type experimental_design: ServiceStatus
        :param data_reports: The status of data_analysis
        :type data_reports: ServiceStatus
        :param model_reports: The status of model reports
        :type model_reports: ServiceStatus
        """
        self._predict             = predict
        self._experimental_design = experimental_design
        self._data_reports        = data_reports
        self._model_reports       = model_reports

    @property
    def predict(self):
        return self._predict

    @predict.setter
    def predict(self, value):
        self._predict = value

    @predict.deleter
    def predict(self):
        self._predict = None

    @property
    def experimental_design(self):
        return self._experimental_design

    @experimental_design.setter
    def experimental_design(self, value):
        self._experimental_design = value

    @experimental_design.deleter
    def experimental_design(self):
        self._experimental_design = None

    @property
    def data_reports(self):
        return self._data_reports

    @data_reports.setter
    def data_reports(self, value):
        self._data_reports = value

    @data_reports.deleter
    def data_reports(self):
        self._data_reports = None

    @property
    def model_reports(self):
        return self._model_reports

    @model_reports.setter
    def model_reports(self, value):
        self._model_reports = value

    @model_reports.deleter
    def model_reports(self):
        self._model_reports = None



from pypif.util.serializable import Serializable

from citrination_client.search.dataset.result.dataset_search_result import DatasetSearchResult


class DatasetMultiSearchResultElement(Serializable):
    """
    A single search result as a part of a dataset multi-query.
    """
    
    def __init__(self, result=None, status=None, **kwargs):
        """
        Constructor.
        
        :param result: A single :class:`DatasetSearchResult` object with the query results.
        :param status: 'SUCCESS', 'ERROR', or 'NOT_EXECUTED'.
        """
        self._result = None
        self.result = result
        self._status = None
        self.status = status

    @property
    def result(self):
        return self._result

    @result.setter
    def result(self, result):
        self._result = self._get_object(DatasetSearchResult, result)

    @result.deleter
    def result(self):
        self._result = None

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        self._status = status

    @status.deleter
    def status(self):
        self._status = None



from citrination_client.search.pif.query.core.base_object_query import BaseObjectQuery
from citrination_client.search.pif.query.core.field_query import FieldQuery


class ClassificationQuery(BaseObjectQuery):
    """
    Class to query against a PIF Classification object.
    """
    
    def __init__(self, logic=None, weight=None, simple=None, simple_weight=None, extract_as=None, extract_all=None,
                 extract_when_missing=None, tags=None, length=None, offset=None, name=None, value=None,
                 query=None, **kwargs):
        """
        Constructor.
        
        :param logic: Logic for this filter. Must be equal to one of "MUST", "MUST_NOT", "SHOULD", or "OPTIONAL".
        :param weight: Weight of the query.
        :param simple: String with the simple query against all fields.
        :param simple_weight: Dictionary of relative paths to their weights for simple queries.
        :param extract_as: String with the alias to save this field under.
        :param extract_all: Boolean setting whether all values in an array should be extracted.
        :param extract_when_missing: Any valid JSON-supported object or PIF object. This value is returned when a value is missing that should be extracted (and the overall query is still satisfied).
        :param tags: One or more :class:`FieldQuery` operations against the tags field.
        :param length: One or more :class:`FieldQuery` operations against the length field.
        :param offset: One or more :class:`FieldQuery` operations against the offset field.
        :param name: One or more :class:`FieldQuery` operations against the name field.
        :param value: One or more :class:`FieldQuery` operations against the value field.
        :param query: One or more :class:`ClassificationQuery` objects for nested queries.
        """
        super(ClassificationQuery, self).__init__(
            logic=logic, weight=weight, simple=simple, simple_weight=simple_weight, extract_as=extract_as,
            extract_all=extract_all, extract_when_missing=extract_when_missing, tags=tags, length=length,
            offset=offset, **kwargs)
        self._name = None
        self.name = name
        self._value = None
        self.value = value
        self._query = None
        self.query = query
    
    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, name):
        self._name = self._get_object(FieldQuery, name)
    
    @name.deleter
    def name(self):
        self._name = None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = self._get_object(FieldQuery, value)

    @value.deleter
    def value(self):
        self._value = None

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, query):
        self._query = self._get_object(ClassificationQuery, query)

    @query.deleter
    def query(self):
        self._query = None



class CitrinationClientError(Exception):

    def __init__(self, message=None, server_response=None):
        if message is not None and server_response is not None:
            message = "{}\nCitrination returned: {}".format(message, server_response)
        super(CitrinationClientError, self).__init__(message)

class APIVersionMismatchException(CitrinationClientError):

    def __init__(self, message="Version mismatch with Citrination identified, server_response=None. Please check for available PyCC updates", server_response=None):
        super(APIVersionMismatchException, self).__init__(message)

class FeatureUnavailableException(CitrinationClientError):

    def __init__(self, message="This feature is unavailable on your Citrination deployment", server_response=None):
        super(FeatureUnavailableException, self).__init__(message)

class UnauthorizedAccessException(CitrinationClientError):

    def __init__(self, message="Access to an unauthorized resource requested", server_response=None):
        super(UnauthorizedAccessException, self).__init__(message)

class ResourceNotFoundException(CitrinationClientError):

    def __init__(self, message="Resource not found", server_response=None):
        super(ResourceNotFoundException, self).__init__(message)

class CitrinationServerErrorException(CitrinationClientError):

    def __init__(self, message=None, server_response=None):
        super(CitrinationServerErrorException, self).__init__(message)

class RequestTimeoutException(CitrinationClientError):

    def __init__(self, message="Request to Citrination host timed out", server_response=None):
        super(RequestTimeoutException, self).__init__(message)

class RateLimitingException(CitrinationClientError):

    def __init__(self, message="Rate limit hit, throttle requests", server_response=None):
        super(RateLimitingException, self).__init__(message)



from citrination_client.models.event import Event
from citrination_client.models.service_status import ServiceStatus
from citrination_client.models.predicted_value import PredictedValue
from citrination_client.models.prediction_result import PredictionResult
from citrination_client.models.projection import Projection
from citrination_client.models.tsne import Tsne
from citrination_client.models.data_view_status import DataViewStatus
from citrination_client.models.columns import *
from citrination_client.models.client import ModelsClient
from citrination_client.models.design import *



from citrination_client.models.design.constraints.base import BaseConstraint

class RealValueConstraint(BaseConstraint):
    """
    Constrains a real valued column to a single value.
    """

    def __init__(self, name, value=None):
        """
        Constructor.

        :param name: The name of the column in the data
            view to which this constraint should be applied
        :type name: str
        :param value: The value the column should be constrained to
        :type value: float
        """
        self._type = "real"        
        self._name = name
        self._value = value

    def options(self):
        return {
            "value": self._value
        }


import yaml
import os
import sys
import citrination_client.util.env as citr_env_vars
from citrination_client.base.errors import CitrinationClientError

home = os.path.expanduser("~")

DEFAULT_CITRINATION_CONFIG_DIRECTORY = ".citrination"
DEFAULT_CITRINATION_CREDENTIALS_FILE = os.path.join(
        home,
        DEFAULT_CITRINATION_CONFIG_DIRECTORY,
        "credentials"
    )

DEFAULT_CITRINATION_PROFILE = "default"
CREDENTIALS_API_KEY_KEY = "api_key"
CREDENTIALS_SITE_KEY = "site"

def load_file_as_yaml(path):
    """
    Given a filepath, loads the file as a dictionary from YAML

    :param path: The path to a YAML file
    """
    with open(path, "r") as f:
      raw_yaml = f.read()
      parsed_dict = yaml.load(raw_yaml)
    return parsed_dict

def get_credentials_from_file(filepath):
    """
    Extracts credentials from the yaml formatted credential filepath
    passed in. Uses the default profile if the CITRINATION_PROFILE env var
    is not set, otherwise looks for a profile with that name in the credentials file.

    :param filepath: The path of the credentials file
    """
    try:
        creds = load_file_as_yaml(filepath)
    except Exception:
        creds = {}

    profile_name = os.environ.get(citr_env_vars.CITRINATION_PROFILE)
    if profile_name is None or len(profile_name) == 0:
        profile_name = DEFAULT_CITRINATION_PROFILE
    api_key = None
    site = None
    try:
        profile = creds[profile_name]
        api_key = profile[CREDENTIALS_API_KEY_KEY]
        site = profile[CREDENTIALS_SITE_KEY]
    except KeyError:
        pass

    return (api_key, site)

def get_preferred_credentials(api_key, site, cred_file=DEFAULT_CITRINATION_CREDENTIALS_FILE):
    """
    Given an API key, a site url and a credentials file path, runs through a prioritized list of credential sources to find credentials.

    Specifically, this method ranks credential priority as follows:
        1. Those passed in as the first two parameters to this method
        2. Those found in the environment as variables
        3. Those found in the credentials file at the profile specified
           by the profile environment variable
        4. Those found in the default stanza in the credentials file

    :param api_key: A Citrination API Key or None
    :param site: A Citrination site URL or None
    :param cred_file: The path to a credentials file
    """
    profile_api_key, profile_site = get_credentials_from_file(cred_file)
    if api_key is None:
        api_key =  os.environ.get(citr_env_vars.CITRINATION_API_KEY)
    if api_key is None or len(api_key) == 0:
        api_key = profile_api_key

    if site is None:
        site = os.environ.get(citr_env_vars.CITRINATION_SITE)
    if site is None or len(site) == 0:
        site = profile_site
    if site is None:
        site = "https://citrination.com"

    return api_key, site



from citrination_client.models.design.constraints.base import BaseConstraint
from citrination_client.base.errors import CitrinationClientError

class ElementalCompositionConstraint(BaseConstraint):
    """
    Constrains a composition or inorganic formula column
    to a particular percentage.
    """

    def __init__(self, name, elements, minimum, maximum):
        """
        Constructor.

        :param name: The name of the column in the data
            view to which this constraint should be applied
        :type name: str
        :param elements: An array of element abbreviations as
            strings, e.g. ["Mg", "C"]
        :type elements: list of str
        :param minimum: The minimum value (<= 100) as a percentage
            at which the specified elements should appear in
            candidate compositions
        :type minimum: float
        :param maximum: The maximum value (<= 100) as a percentage
            at which the specified elements should appear in
            candidate compositions
        :type maximum: float
        """
        if not 0 <= minimum <= 100:
            raise CitrinationClientError("ElementalCompositionConstraint requires that minimum be between 0 and 100")

        if not 0 <= maximum <= 100:
            raise CitrinationClientError("ElementalCompositionConstraint requires that maximum be between 0 and 100")

        if not maximum >= minimum:
            raise CitrinationClientError("ElementalCompositionConstraint requires that maximum be greater than minimum")


        self._type = "elementalCompositionConstraint"
        self._elements = elements
        self._name = name
        self._min = minimum
        self._max = maximum

    def options(self):
        return {
            "elements": self._elements,
            "min": self._min,
            "max": self._max
        }


class Dataset(object):
    """
    Class representation of a dataset on Citrination.
    """

    def __init__(self, id, name=None, description=None,
            created_at=None):
        """
        Constructor.

        :param id: The ID of the dataset (required for instantiation)
        :type id: int
        :param name: The name of the dataset
        :type name: str
        :param description: The description of the dataset
        :type description: str
        :param created_at: The timestamp for creation of the dataset
        :type created_at: str
        """
        self._name = name
        self._description = description
        self._id = id
        self._created_at = created_at

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @name.deleter
    def name(self):
        self._name = None

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @description.deleter
    def description(self):
        self._description = None

    @property
    def created_at(self):
        return self._created_at

    @created_at.setter
    def created_at(self, value):
        self._created_at = value

    @created_at.deleter
    def created_at(self):
        self._created_at = None



class DatasetVersion(object):
    """
    Class representation of a version of a dataset on Citrination.
    """

    def __init__(self, number):
        """
        Constructor.

        :param number: The number of the dataset version
        :type number: ints
        """
        self._number = number

    @property
    def number(self):
        return self._number

    @number.setter
    def number(self, value):
        self._number = value

    @number.deleter
    def number(self):
        self._number = None


class Tsne(object):
    """
    A TSNE analysis which can be extracted from the data analysis
    for a trained model on Citrination.
    """

    def __init__(self):
        """
        Constructor
        """
        self._projections = {}

    def add_projection(self, key, projection):
        """
        Register a projection under a descriptor key.

        :param key: The descriptor key for the projection
        :type key: str
        :param projection: The projection for the provided descriptor key
        :type projection: :class:`Projection`
        """
        self._projections[key] = projection

    def projections(self):
        """
        List the descriptor keys with registered projections.

        :return: List of descriptor keys
        """
        return self._projections.keys()

    def get_projection(self, key):
        """
        Retrieves the projection registered under a particular
        descriptor key.

        :param key: A descriptor key
        :return: A :class:`Projection`
        """
        return self._projections.get(key)


from citrination_client.models.design.constraints.base import BaseConstraint

class CategoricalConstraint(BaseConstraint):
    """
    Constrains a column to a particular set of categorical
    values.
    """

    def __init__(self, name, accepted_categories):
        """
        Constructor.

        :param name: The name of the column in the data
            view to which this constraint should be applied
        :type name: str
        :param accepted_categories: An array of categories to constrain the name to
        :type accepted_categories: list of str
        """
        self._type = "categorical"
        self._name = name
        self._categories = accepted_categories

    def options(self):
        return {
            "categories": self._categories,
        }


from citrination_client.base.base_client import BaseClient
from citrination_client.base.errors import *
from citrination_client.data import *
from citrination_client.data import routes as routes

from pypif import pif

import os
import shutil
import requests

class DataClient(BaseClient):
    """
    Client encapsulating data management behavior.
    """

    def __init__(self, api_key, host="https://citrination.com", suppress_warnings=False):
        """
        Constructor.

        :param api_key: A users API key, as a string
        :type api_key: str
        :param host: The base URL of the citrination site, e.g. https://citrination.com
        :type host: str
        :param suppress_warnings: Whether or not usage warnings should be
            printed to stdout
        :type suppress_warnings: bool
        """
        members = [
            "upload",
            "list_files",
            "matched_file_count",
            "get_dataset_files",
            "get_dataset_file",
            "download_files",
            "create_dataset",
            "create_dataset_version"
        ]
        super(DataClient, self).__init__(api_key, host, members, suppress_warnings=suppress_warnings)

    def upload(self, dataset_id, source_path, dest_path=None):
        """
        Upload a file, specifying source and dest paths a file (acts as the scp command).asdfasdf

        :param source_path: The path to the file on the source host asdf
        :type source_path: str
        :param dest_path: The path to the file where the contents of the upload will be written (on the dest host)
        :type dest_path: str
        :return: The result of the upload process
        :rtype: :class:`UploadResult`
        """
        upload_result = UploadResult()
        source_path = str(source_path)
        if not dest_path:
            dest_path = source_path
        else:
            dest_path = str(dest_path)
        if os.path.isdir(source_path):
            for path, subdirs, files in os.walk(source_path):
                for name in files:
                    path_without_root_dir = path.split("/")[-1:] + [name]
                    current_dest_path = os.path.join(dest_path, *path_without_root_dir)
                    current_source_path = os.path.join(path, name)
                    try:
                        if self.upload(dataset_id, current_source_path, current_dest_path).successful():
                            upload_result.add_success(current_source_path)
                        else:
                            upload_result.add_failure(current_source_path,"Upload failure")
                    except (CitrinationClientError, ValueError) as e:
                        upload_result.add_failure(current_source_path, e.message)
            return upload_result
        elif os.path.isfile(source_path):
            file_data = { "dest_path": str(dest_path), "src_path": str(source_path)}
            j = self._get_success_json(self._post_json(routes.upload_to_dataset(dataset_id), data=file_data))
            s3url = _get_s3_presigned_url(j)
            with open(source_path, 'rb') as f:
                r = requests.put(s3url, data=f, headers=j["required_headers"])
                if r.status_code == 200:
                    data = {'s3object': j['url']['path'], 's3bucket': j['bucket']}
                    self._post_json(routes.update_file(j['file_id']), data=data)
                    upload_result.add_success(source_path)
                    return upload_result
                else:
                    raise CitrinationClientError("Failure to upload {} to Citrination".format(source_path))
        else:
            raise ValueError("No file at specified path {}".format(source_path))

    def list_files(self, dataset_id, glob=".", is_dir=False):
        """
        List matched filenames in a dataset on Citrination.

        :param dataset_id: The ID of the dataset to search for files.
        :type dataset_id: int
        :param glob: A pattern which will be matched against files in the dataset.
        :type glob: str
        :param is_dir: A boolean indicating whether or not the pattern should match against the beginning of paths in the dataset.
        :type is_dir: bool
        :return: A list of filepaths in the dataset matching the provided glob.
        :rtype: list of strings
        """
        data = {
            "list": {
                "glob": glob,
                "isDir": is_dir
            }
        }
        return self._get_success_json(self._post_json(routes.list_files(dataset_id), data, failure_message="Failed to list files for dataset {}".format(dataset_id)))['files']

    def matched_file_count(self, dataset_id, glob=".", is_dir=False):
        """
        Returns the number of files matching a pattern in a dataset.

        :param dataset_id: The ID of the dataset to search for files.
        :type dataset_id: int
        :param glob: A pattern which will be matched against files in the dataset.
        :type glob: str
        :param is_dir: A boolean indicating whether or not the pattern should match against the beginning of paths in the dataset.
        :type is_dir: bool
        :return: The number of matching files
        :rtype: int
        """
        list_result = self.list_files(dataset_id, glob, is_dir)
        return len(list_result)

    def get_dataset_files(self, dataset_id, glob=".", is_dir=False, version_number=None):
        """
        Retrieves URLs for the files matched by a glob or a path to a directory
        in a given dataset.

        :param dataset_id: The id of the dataset to retrieve files from
        :type dataset_id: int
        :param glob: A regex used to select one or more files in the dataset
        :type glob: str
        :param is_dir: Whether or not the supplied pattern should be treated as a directory to search in
        :type is_dir: bool
        :param version_number: The version number of the dataset to retrieve files from
        :type version_number: int
        :return: A list of dataset files whose paths match the provided pattern.
        :rtype: list of :class:`DatasetFile`
        """
        if version_number is None:
            latest = True
        else:
            latest = False

        data = {
            "download_request": {
                "glob": glob,
                "isDir": is_dir,
                "latest": latest
            }
        }

        failure_message = "Failed to get matched files in dataset {}".format(dataset_id)

        versions = self._get_success_json(self._post_json(routes.matched_files(dataset_id), data, failure_message=failure_message))['versions']

        # if you don't provide a version number, only the latest
        # will be included in the response body
        if version_number is None:
            version = versions[0]
        else:
            try:
                version = list(filter(lambda v: v['number'] == version_number, versions))[0]
            except IndexError:
                raise ResourceNotFoundException()

        return list(
            map(
                lambda f: DatasetFile(path=f['filename'], url=f['url']), version['files']
                )
            )

    def get_dataset_file(self, dataset_id, file_path, version = None):
        """
        Retrieves a dataset file matching a provided file path

        :param dataset_id: The id of the dataset to retrieve file from
        :type dataset_id: int
        :param file_path: The file path within the dataset
        :type file_path: str
        :param version: The dataset version to look for the file in. If nothing is supplied, the latest dataset version will be searched
        :type version: int
        :return: A dataset file matching the filepath provided
        :rtype: :class:`DatasetFile`
        """
        return self.get_dataset_files(dataset_id, "^{}$".format(file_path), version_number=version)[0]

    def download_files(self, dataset_files, destination='.'):
        """
        Downloads file(s) to a local destination.

        :param dataset_files:
        :type dataset_files: list of :class: `DatasetFile`
        :param destination: The path to the desired local download destination
        :type destination: str
        :param chunk: Whether or not to chunk the file. Default True
        :type chunk: bool
        """
        if not isinstance(dataset_files, list):
            dataset_files = [dataset_files]

        for f in dataset_files:
            filename = f.path.lstrip('/')
            local_path = os.path.join(destination, filename)

            if not os.path.isdir(os.path.dirname(local_path)):
                os.makedirs(os.path.dirname(local_path))

            r = requests.get(f.url, stream=True)

            with open(local_path, 'wb') as output_file:
                shutil.copyfileobj(r.raw, output_file)

    def get_pif(self, dataset_id, uid, dataset_version = None):
        """
        Retrieves a PIF from a given dataset.

        :param dataset_id: The id of the dataset to retrieve PIF from
        :type dataset_id: int
        :param uid: The uid of the PIF to retrieve
        :type uid: str
        :param dataset_version: The dataset version to look for the PIF in. If nothing is supplied, the latest dataset version will be searched
        :type dataset_version: int
        :return: A :class:`Pif` object
        :rtype: :class:`Pif`
        """
        failure_message = "An error occurred retrieving PIF {}".format(uid)
        if dataset_version == None:
            response = self._get(routes.pif_dataset_uid(dataset_id, uid), failure_message=failure_message)
        else:
            response = self._get(routes.pif_dataset_version_uid(dataset_id, uid, dataset_version), failure_message=failure_message)

        return pif.loads(response.content.decode("utf-8"))

    def create_dataset(self, name=None, description=None, public=False):
        """
        Create a new data set.

        :param name: name of the dataset
        :type name: str
        :param description: description for the dataset
        :type description: str
        :param public: A boolean indicating whether or not the dataset should be public.
        :type public: bool
        :return: The newly created dataset.
        :rtype: :class:`Dataset`
        """
        data = {
            "public": _convert_bool_to_public_value(public)
        }
        if name:
            data["name"] = name
        if description:
            data["description"] = description
        dataset = {"dataset": data}
        failure_message = "Unable to create dataset"
        result = self._get_success_json(self._post_json(routes.create_dataset(), dataset, failure_message=failure_message))

        return _dataset_from_response_dict(result)

    def update_dataset(self, dataset_id, name=None, description=None, public=None):
        """
        Update a data set.

        :param dataset_id: The ID of the dataset to update
        :type dataset_id: int
        :param name: name of the dataset
        :type name: str
        :param description: description for the dataset
        :type description: str
        :param public: A boolean indicating whether or not the dataset should
            be public.
        :type public: bool
        :return: The updated dataset.
        :rtype: :class:`Dataset`
        """
        data = {
            "public": _convert_bool_to_public_value(public)
        }

        if name:
            data["name"] = name
        if description:
            data["description"] = description

        dataset = {"dataset": data}
        failure_message = "Failed to update dataset {}".format(dataset_id)
        response = self._get_success_json(self._post_json(routes.update_dataset(dataset_id), data=dataset, failure_message=failure_message))

        return _dataset_from_response_dict(response)

    def create_dataset_version(self, dataset_id):
        """
        Create a new data set version.

        :param dataset_id: The ID of the dataset for which the version must be bumped.
        :type dataset_id: int
        :return: The new dataset version.
        :rtype: :class:`DatasetVersion`
        """
        failure_message = "Failed to create dataset version for dataset {}".format(dataset_id)
        number = self._get_success_json(self._post_json(routes.create_dataset_version(dataset_id), data={}, failure_message=failure_message))['dataset_scoped_id']

        return DatasetVersion(number=number)

def _dataset_from_response_dict(dataset):
    return Dataset(dataset['id'], name=dataset['name'],
        description=dataset['description'], created_at=dataset['created_at'])

def _convert_bool_to_public_value(val):
    if val == None:
        return None
    if val == False:
        return '0'
    if val == True:
        return '1'
    # for backwards compatability, support the old API #utahisrad
    if val == '0' or val == '1':
        return val

def _get_s3_presigned_url(response_dict):
    """
    Helper method to create an S3 presigned url from the response dictionary.
    """
    url = response_dict['url']
    return url['scheme']+'://'+url['host']+url['path']+'?'+url['query']



from citrination_client.base.base_client import BaseClient
from citrination_client.models import *
from citrination_client.models.design import *
from citrination_client.models import routes as routes
from citrination_client.base.errors import CitrinationClientError
from citrination_client.data import Dataset
from citrination_client.models.data_view import DataView
from citrination_client.models.columns.column_factory import ColumnFactory

import time

class ModelsClient(BaseClient):
    """
    A client that encapsulates interactions with models on Citrination.
    """

    def __init__(self, api_key, webserver_host="https://citrination.com", suppress_warnings=False):
        members = [
            "tsne",
            "predict"
        ]
        super(ModelsClient, self).__init__(api_key, webserver_host, members, suppress_warnings=suppress_warnings)

    def tsne(self, data_view_id):
        """
        Get the t-SNE projection, including responses and tags.

        :param data_view_id: The ID of the data view to retrieve TSNE from
        :type data_view_id: int
        :return: The TSNE analysis
        :rtype: :class:`Tsne`
        """
        analysis = self._data_analysis(data_view_id)
        projections = analysis['projections']
        tsne = Tsne()
        for k, v in projections.items():
            projection = Projection(
                xs=v['x'],
                ys=v['y'],
                responses=v['label'],
                tags=v['inputs'],
                uids=v['uid']
            )
            tsne.add_projection(k, projection)

        return tsne

    def predict(self, data_view_id, candidates, method="scalar", use_prior=True):
        """
        Predict endpoint

        :param data_view_id: The ID of the data view to use for prediction
        :type data_view_id: str
        :param candidates: A list of candidates to make predictions on
        :type candidates: list of dicts
        :param method: Method for propagating predictions through model
            graphs
        :type method: str ("scalar" or "from_distribution")
        :param use_prior:  Whether to apply prior values implied by the property descriptors
        :type use_prior: bool
        :return: The results of the prediction
        :rtype: list of :class:`PredictionResult`
        """
        body = self._get_predict_body(candidates, method, use_prior)
        failure_message = "Error while making prediction for data view {}".format(data_view_id)
        response_dict = self._get_success_json(
            self._post_json(routes.data_view_predict(data_view_id), data=body, failure_message=failure_message))
        candidate_dicts = response_dict["candidates"]
        return list(
            map(
                lambda c: _get_prediction_result_from_candidate(c), candidate_dicts
            )
        )

    def _data_analysis(self, data_view_id):
        """
        Data analysis endpoint.

        :param data_view_id: The model identifier (id number for data views)
        :type data_view_id: str
        :return: dictionary containing information about the data, e.g. dCorr and tsne
        """
        failure_message = "Error while retrieving data analysis for data view {}".format(data_view_id)
        return self._get_success_json(self._get(routes.data_analysis(data_view_id), failure_message=failure_message))

    def _get_predict_body(self, candidates, method="scalar", use_prior=True):
        if not (method == "scalar" or method == "from_distribution"):
            raise ValueError("{} method not supported".format(method))

        # If a single candidate is passed, wrap in a list for the user
        if not isinstance(candidates, list):
            candidates = [candidates]

        return {
            "predictionRequest": {
                "predictionSource": method,
                "usePrior":         use_prior,
                "candidates":       candidates
            }
        }

    def submit_design_run(self, data_view_id, num_candidates, effort, target=None, constraints=[], sampler="Default"):
        """
        Submits a new experimental design run

        :param data_view_id: The ID number of the data view to which the
            run belongs, as a string
        :type data_view_id: str
        :param num_candidates: The number of candidates to return
        :type num_candidates: int
        :param target: An :class:``Target`` instance representing
            the design run optimization target
        :type target: :class:``Target``
        :param constraints: An array of design constraints (instances of
            objects which extend :class:``BaseConstraint``)
        :type constraints: list of :class:``BaseConstraint``
        :param sampler: The name of the sampler to use during the design run:
            either "Default" or "This view"
        :type sampler: str
        :return: A :class:`DesignRun` instance containing the UID of the
            new run
        """

        if effort > 30:
            raise CitrinationClientError("Parameter effort must be less than 30 to trigger a design run")

        if target is not None:
            target = target.to_dict()

        constraint_dicts = [c.to_dict() for c in constraints]

        body = {
            "num_candidates": num_candidates,
            "target":         target,
            "effort":         effort,
            "constraints":    constraint_dicts,
            "sampler":        sampler
        }

        url = routes.submit_data_view_design(data_view_id)

        response = self._post_json(url, body).json()

        return DesignRun(response["data"]["design_run"]["uid"])

    def get_design_run_status(self, data_view_id, run_uuid):
        """
        Retrieves the status of an in progress or completed design run

        :param data_view_id: The ID number of the data view to which the
            run belongs, as a string
        :type data_view_id: str
        :param run_uuid: The UUID of the design run to retrieve status for
        :type run_uuid: str
        :return: A :class:`ProcessStatus` object
        """

        url = routes.get_data_view_design_status(data_view_id, run_uuid)

        response = self._get(url).json()

        status = response["data"]

        return ProcessStatus(
            result=status.get("result"),
            progress=status.get("progress"),
            status=status.get("status"),
            messages=status.get("messages")
        )

    def get_design_run_results(self, data_view_id, run_uuid):
        """
        Retrieves the results of an existing designrun

        :param data_view_id: The ID number of the data view to which the
            run belongs, as a string
        :type data_view_id: str
        :param run_uuid: The UUID of the design run to retrieve results from
        :type run_uuid: str
        :return: A :class:`DesignResults` object
        """

        url = routes.get_data_view_design_results(data_view_id, run_uuid)

        response = self._get(url).json()

        result = response["data"]

        return DesignResults(
            best_materials=result.get("best_material_results"),
            next_experiments=result.get("next_experiment_results")
        )

    def get_data_view(self, data_view_id):
        """
        Retrieves a summary of information for a given data view
            - view id
            - name
            - description
            - columns

        :param data_view_id: The ID number of the data view to which the
            run belongs, as a string
        :type data_view_id: str
        """

        url = routes.get_data_view(data_view_id)

        response = self._get(url).json()

        result = response["data"]["data_view"]

        datasets_list = []
        for dataset in result["datasets"]:
            datasets_list.append(Dataset(
                name=dataset["name"],
                id=dataset["id"],
                description=dataset["description"]
            ))

        columns_list = []
        for column in result["columns"]:
            columns_list.append(ColumnFactory.from_dict(column))

        return DataView(
            view_id=data_view_id,
            name=result["name"],
            description=result["description"],
            datasets=datasets_list,
            columns=columns_list,
        )

    def kill_design_run(self, data_view_id, run_uuid):
        """
        Kills an in progress experimental design run

        :param data_view_id: The ID number of the data view to which the
            run belongs, as a string
        :type data_view_id: str
        :param run_uuid: The UUID of the design run to kill
        :type run_uuid: str
        :return: The UUID of the design run
        """

        url = routes.kill_data_view_design_run(data_view_id, run_uuid)

        response = self._delete(url).json()
        return response["data"]["uid"]

    def get_data_view_service_status(self, data_view_id):
        """
        Retrieves the status for all of the services associated with a data view:
            - predict
            - experimental_design
            - data_reports
            - model_reports

        :param data_view_id: The ID number of the data view to which the
            run belongs, as a string
        :type data_view_id: str
        :return: A :class:`DataViewStatus`
        :rtype: DataViewStatus
        """

        url = routes.get_data_view_status(data_view_id)

        response = self._get(url).json()

        result = response["data"]["status"]

        return DataViewStatus(
            predict = ServiceStatus.from_response_dict(result["predict"]),
            experimental_design = ServiceStatus.from_response_dict(result["experimental_design"]),
            data_reports = ServiceStatus.from_response_dict(result["data_reports"]),
            model_reports = ServiceStatus.from_response_dict(result["model_reports"])
        )

def _get_prediction_result_from_candidate(candidate_dict):
    result = PredictionResult()
    for k, v in candidate_dict.items():
        result.add_value(k, PredictedValue(k, v[0], v[1]))

    return result



from pypif.util.serializable import Serializable

from citrination_client.search.core.query.filter import Filter


class FileQuery(Serializable):
    """
    Class to store information about a file query.
    """

    def __init__(self, logic=None, weight=None, simple=None, simple_weight=None, id=None, name=None, content=None,
                 updated_at=None, query=None, **kwargs):
        """
        Constructor.

        :param logic: The logic to apply to the query ('SHOULD', 'MUST', 'MUST_NOT', or 'OPTIONAL').
        :param weight: Weight for the query.
        :param simple: String with the simple search to run against all fields.
        :param simple_weight: Dictionary of relative paths to their weights for simple queries.
        :param id: One or more :class:`Filter` objects with filters against the id field.
        :param name: One or more :class:`Filter` objects with filters against the name field.
        :param content: One or more :class:`Filter` objects with filters against the content field.
        :param updated_at: One or more :class:`Filter` objects with filters against the time that the dataset was last updated.
        :param query: One or more :class:`DatasetQuery` objects with nested queries.
        """
        self._logic = None
        self.logic = logic
        self._weight = None
        self.weight = weight
        self._simple = None
        self.simple = simple
        self._simple_weight = None
        self.simple_weight = simple_weight
        self._id = None
        self.id = id
        self._name = None
        self.name = name
        self._content = None
        self.content = content
        self._updated_at = None
        self.updated_at = updated_at
        self._query = None
        self.query = query

    @property
    def logic(self):
        return self._logic

    @logic.setter
    def logic(self, logic):
        self._logic = logic

    @logic.deleter
    def logic(self):
        self._logic = None

    @property
    def weight(self):
        return self._weight

    @weight.setter
    def weight(self, weight):
        self._weight = weight

    @weight.deleter
    def weight(self):
        self._weight = None

    @property
    def simple(self):
        return self._simple

    @simple.setter
    def simple(self, simple):
        self._simple = simple

    @simple.deleter
    def simple(self):
        self._simple = None

    @property
    def simple_weight(self):
        return self._simple_weight

    @simple_weight.setter
    def simple_weight(self, simple_weight):
        self._simple_weight = simple_weight

    @simple_weight.deleter
    def simple_weight(self):
        self._simple_weight = None

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, id):
        self._id = self._get_object(Filter, id)

    @id.deleter
    def id(self):
        self._id = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = self._get_object(Filter, name)

    @name.deleter
    def name(self):
        self._name = None

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, content):
        self._content = self._get_object(Filter, content)

    @content.deleter
    def content(self):
        self._content = None

    @property
    def updated_at(self):
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at):
        self._updated_at = self._get_object(Filter, updated_at)

    @updated_at.deleter
    def updated_at(self):
        self._updated_at = None

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, query):
        self._query = self._get_object(FileQuery, query)

    @query.deleter
    def query(self):
        self._query = None



from citrination_client.models.columns.base import BaseColumn

class AlloyCompositionColumn(BaseColumn):
    """
    An alloy composition column configuration for a data view. Parameterized
    with the basic column options, plus the balance element for the column
    and the basis value for the composition.
    """

    TYPE = "Alloy composition"

    def __init__(self, name, role, balance_element, group_by_key=False, units=None, basis=100.0):
        """
        Constructor.

        :param name: The name of the column
        :type name: str
        :param role: The role the column will play in machine learning:
                       "Input"
                       "Output"
                       "Latent Variable"
                       "Ignore"
        :type role: str
        :param group_by_key: Whether or not this column should be used for
            grouping during cross validation
        :type group_by_key: bool
        :param units: Optionally, the units for the column
        :type units: str
        :param balance_element: The element making up the balance in the
            composition
        :type balance_element: str
        :param basis: The total amount of composition when deciding how to fill
            the balance
        :type basis: float

        """
        super(AlloyCompositionColumn, self).__init__(name=name,
                                                role=role,
                                                group_by_key=group_by_key,
                                                units=units)
        self._balance_element = balance_element
        self._basis = basis

    def build_options(self):
        return {
            "balance_element": self.balance_element,
            "basis": self.basis
        }

    @property
    def basis(self):
        return self._basis

    @basis.setter
    def basis(self, value):
        self._basis = value

    @basis.deleter
    def basis(self):
        self._basis = None

    @property
    def balance_element(self):
        return self._balance_element

    @balance_element.setter
    def balance_element(self, value):
        self._balance_element = value

    @balance_element.deleter
    def balance_element(self):
        self._balance_element = None



class DesignResults(object):
    """
    The result of a design run. Consists of a set of dictionaries representing
    the candidates which are the best materials according to the target
    and the constraints and a set of dictionaries representing candidates
    which would be the most useful in improving the quality of the model.
    """

    def __init__(self, best_materials, next_experiments):
        """
        Constructor.

        :param best_materials: An array of candidate dictionaries
        :type best_materials: list of dictionaries
        :param next_experiments: An array of candidate dictionaries
        :type next_experiments: list of dictionaries
        """
        self._best_materials = best_materials
        self._next_experiments = next_experiments

    @property
    def best_materials(self):
        return self._best_materials

    @best_materials.setter
    def best_materials(self, value):
        self._best_materials = value

    @best_materials.deleter
    def best_materials(self):
        self._best_materials = None

    @property
    def next_experiments(self):
        return self._next_experiments

    @next_experiments.setter
    def next_experiments(self, value):
        self._next_experiments = value

    @next_experiments.deleter
    def next_experiments(self):
        self._next_experiments = None


from citrination_client.search.core.query.data_scope import DataScope


class BaseReturningQuery(DataScope):
    """
    Base class for all queries against datasets and the items that they contain on Citrination.
    """

    def __init__(self, query=None, extraction_sort=None, from_index=None, size=None, random_results=None,
                 random_seed=None, score_relevance=None, return_max_score=None, timeout=None, **kwargs):
        """
        Base class for all queries against datasets and the items that they contain on Citrination.

        :param query: One or more :class:`DataQuery` objects with the queries to run.
        :param extraction_sort: A single :class:`ExtractionSort` object for sorting.
        :param from_index: Index of the first hit that should be returned.
        :param size: Total number of hits the should be returned.
        :param random_results: Whether to return a random set of records.
        :param random_seed: The random seed to use.
        :param score_relevance: Whether to use relevance scoring.
        :param return_max_score: Whether to return the maximum score.
        :param timeout: The number of milliseconds to wait for the query to execute.
        """
        super(BaseReturningQuery, self).__init__(query=query, extraction_sort=extraction_sort, **kwargs)
        if 'from' in 'kwargs':
            self.from_index = kwargs['from']
        self._from = None
        self.from_index = from_index
        self._size = None
        self.size = size
        self._random_results = None
        self.random_results = random_results
        self._random_seed = None
        self.random_seed = random_seed
        self._score_relevance = None
        self.score_relevance = score_relevance
        self._return_max_score = None
        self.return_max_score = return_max_score
        self._timeout = None
        self.timeout = timeout

    @property
    def from_index(self):
        return self._from

    @from_index.setter
    def from_index(self, from_index):
        self._from = from_index

    @from_index.deleter
    def from_index(self):
        self._from = None

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, size):
        self._size = size

    @size.deleter
    def size(self):
        self._size = None

    @property
    def random_results(self):
        return self._random_results

    @random_results.setter
    def random_results(self, random_results):
        self._random_results = random_results

    @random_results.deleter
    def random_results(self):
        self._random_results = None

    @property
    def random_seed(self):
        return self._random_seed

    @random_seed.setter
    def random_seed(self, random_seed):
        self._random_seed = random_seed

    @random_seed.deleter
    def random_seed(self):
        self._random_seed = None

    @property
    def score_relevance(self):
        return self._score_relevance

    @score_relevance.setter
    def score_relevance(self, score_relevance):
        self._score_relevance = score_relevance

    @score_relevance.deleter
    def score_relevance(self):
        self._score_relevance = None

    @property
    def return_max_score(self):
        return self._return_max_score

    @return_max_score.setter
    def return_max_score(self, return_max_score):
        self._return_max_score = return_max_score

    @return_max_score.deleter
    def return_max_score(self):
        self._return_max_score = None

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        self._timeout = timeout

    @timeout.deleter
    def timeout(self):
        self._timeout = None



class PredictionResult(object):
    """
    The collection of predicted values resulting from a prediction.
    """

    def __init__(self):
        self._values = {}

    def add_value(self, key, value):
        """
        Registers a predicted value in the result.

        :param key: The descriptor key for the predicted value
        :type key: str
        :param value: A :class:`PredictedValue`
        :type value: object
        :return: None
        :rtype: NoneType
        """
        self._values[key] = value

    def get_value(self, key):
        """
        Retrieves a predicted value.

        :param key: A descriptor key for a registered predicted value.
        :type key: str
        :return: The value stored at the provided descriptor key. None if no key is provided.
        :rtype: :class:`PredictedValue`
        """
        try:
            return self._values[key]
        except KeyError:
            return None

    def all_keys(self):
        """
        Retrieves a list of all the values which were predicted.

        :return: A list of keys for which predictions have been made and can
            be retrieved using `get_value`
        :rtype: list of str
        """

        return self._values.keys()